
// The D/Objective-C Bridge
// Copyright © 2007  Michel Fortin
// <http://michelf.com/d-objc-bridge/>
//
// This program is free software: you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free 
// Software Foundation, either version 3 of the License, or (at your option) 
// any later version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT 
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for 
// more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program. If not, see <http://www.gnu.org/licenses/>.

module cocoa.foundation.dictionary;

import cocoa.foundation.object;
import cocoa.foundation.string;
import cocoa.foundation.array;
import cocoa.foundation.enumerator;

import objc.bridge;
import objc.runtime;


/**
 * Create an NSDictionary from an associative array.
 */
NSDictionary toNSDictionary(Object[Object] objects) {
	return new NSDictionary(objects);
}


/**
 * Wrapper for Objective-C class NSDictionary.
 *
 * Subclassing notes: NSDictionary is a cluster class. If you subclass it, 
 * you need to create your own Objective-C instance first with objcAlloc and 
 * pass it as the argument for the NSDictionary constructor. Otherwise, if 
 * you call another NSDictionary constructor which creates the Objective-C 
 * instance for you, that instance will not be a capsule for your subclass.
 */
class NSDictionary : NSObject /*, NSCoding */ {
	mixin ObjcClusterWrap;
	this(objc.runtime.id instance) { super(instance); }

	uint length() {
		return invokeObjcSuper!(uint, "count")();
	}

	NSEnumerator keyEnumerator() {
		return invokeObjcSuper!(NSEnumerator, "keyEnumerator")();
	}
	
	Object opIndex(Object key) {
		return invokeObjcSuper!(Object, "objectForKey:")(key);
	}
	
	Object opIndex(char[] key) {
		return opIndex(key.toNSString());
	}
	
	mixin ObjcBindMethod!(length, uint, "count");
	mixin ObjcBindMethod!(keyEnumerator, NSEnumerator, "keyEnumerator");
	mixin ObjcBindMethod!(opIndex, Object, "objectForKey:", Object);


	// NSExtendedDictionary

	NSArray keys() {
		return invokeObjcSuper!(NSArray, "allKeys")();
	}

	NSArray keysForObject(Object object) {
		return invokeObjcSuper!(NSArray, "allKeysForObject:")(object);
	}

	NSArray values() {
		return invokeObjcSuper!(NSArray, "allValues")();
	}
	
	NSString toStringsFileFormat() {
		return invokeObjcSuper!(NSString, "descriptionInStringsFileFormat")();
	}
	
	bool isEqualToDictionary(NSDictionary otherDictionary) {
		return invokeObjcSuper!(bool, "isEqualToDictionary:")(otherDictionary);
	}
	
	uint opEquals(Object[Object] objects) {
		if (this.length != objects.length)  return false;
		if (length == 0)                    return true;
		return isEqualToDictionary(objects.toNSDictionary());
	}

	NSEnumerator objectEnumerator() {
		return invokeObjcSuper!(NSEnumerator, "objectEnumerator")();
	}

	NSArray objectsForKeys(NSArray keys, Object marker) {
		return invokeObjcSuper!(NSArray, "objectsForKeys:notFoundMarker:")(keys, marker);
	}

	bool writeToFile(NSString path, bool useAuxiliaryFile) {
		return invokeObjcSuper!(bool, "writeToFile:atomically:")(path, useAuxiliaryFile);
	}

//	bool writeToURL(NSURL url, bool useAuxiliaryFile) {
//		mixin ObjcMethod!(bool, "writeToURL:atomically:", NSURL, bool);
//		mixin ObjcOverride!(writeToFile);
//		return invokeObjcSuper!(path, useAuxiliaryFile);
//	}
	
	alias NSObject.opEquals opEquals; // also import NSObject opEquals.
	
	mixin ObjcBindMethod!(keys, NSArray, "allKeys");
	mixin ObjcBindMethod!(keysForObject, NSArray, "allKeysForObject:", Object);
	mixin ObjcBindMethod!(values, NSArray, "allValues");
	mixin ObjcBindMethod!(toStringsFileFormat, NSString, "descriptionInStringsFileFormat");
	mixin ObjcBindMethod!(isEqualToDictionary, bool, "isEqualToDictionary:", NSDictionary);
	mixin ObjcBindMethod!(objectEnumerator, NSEnumerator, "objectEnumerator");
	mixin ObjcBindMethod!(objectsForKeys, NSArray, "objectsForKeys:notFoundMarker:", NSArray, Object);
	mixin ObjcBindMethod!(writeToFile, bool, "writeToFile:atomically:", NSString, bool);
//	mixin ObjcBindMethod!(writeToURL, bool, "writeToURL:atomically:", NSURL, bool);


	// NSDictionaryCreation

	this() {
		super(objcClusterAlloc);
		objcInstance = invokeObjcSuper!(id, "init")();
	}

	this(NSString path) {
		this(objcClusterAlloc);
		objcInstance = invokeObjcSuper!(id, "initWithContentOfFile:")(path);
	}

//	this(NSURL url) {
//		this(objcClusterAlloc);
//		objcInstance = invokeObjcSuper!(id, "initWithContentOfURL:")(url);
//	}

	this(Object[Object] aa) {
		this(aa.values.toObjcInstances(), aa.keys.toObjcInstances());
	}

	this(Object[char[]] aa) {
		this(aa.values.toObjcInstances(), aa.keys.toObjcInstances());
	}

	this(Object[] objects, Object[] keys)
	in { assert(objects.length == keys.length); }
	body {
		this(objects.toObjcInstances(), keys.toObjcInstances());
	}

	this(objc.runtime.id[] objects, objc.runtime.id[] keys) 
	in { assert(objects.length == keys.length); }
	body {
		this(objcClusterAlloc);
		objcInstance = invokeObjcSuper!(id, "initWithObjects:forKeys:count:")(objects.ptr, keys.ptr, objects.length);
	}
	
	this(NSArray objects, NSArray keys) {
		this(objcClusterAlloc);
		objcInstance = invokeObjcSuper!(id, "initWithObjects:forKeys:")(objects, keys);
	}
	
	this(NSDictionary otherDictionary) {
		this(objcClusterAlloc);
		objcInstance = invokeObjcSuper!(id, "initWithDictionary:")(otherDictionary);
	}

	this(NSDictionary otherDictionary, bool copyItems) {
		this(objcClusterAlloc);
		objcInstance = invokeObjcSuper!(id, "initWithDictionary:copyItems:")(otherDictionary, copyItems);
	}
	
	
	// Other goodies to work like a D associative array.

	alias opIndex opIn_r;

	int opApply(int delegate(ref Object) dg) {
		auto e = objectEnumerator;
		Object obj;
		while ((obj = e.nextObject) !is null) {
			int broken = dg(obj);
			if (broken) return broken;
		}
		return 0;
	}
	
	int opApply(int delegate(ref Object, ref Object) dg) {
		auto e = keyEnumerator;
		Object key;
		while ((key = e.nextObject) !is null) {
			Object obj = this[key];
			int broken = dg(key, obj);
			if (broken) return broken;
		}
		return 0;
	}
}


/**
 * Wrapper for Objective-C class NSMutableDictionary.
 *
 * Subclassing notes: NSMutableDictionary is a cluster class. If you subclass 
 * it, you need to create your own Objective-C instance first with objcAlloc 
 * and pass it as the argument for the NSMutableDictionary constructor. 
 * Otherwise, if you call another NSMutableDictionary constructor which 
 * creates the Objective-C instance for you, that instance will not be a capsule for your subclass.
 */
class NSMutableDictionary : NSDictionary {
	mixin ObjcClusterWrap;
	this(objc.runtime.id instance) { super(instance); }

	/** Remove element at key. */
	void remove(Object key) {
		return invokeObjcSuper!(void, "removeObjectForKey:")(key);
	}
	/** ditto */
	final void remove(char[] key) {
		remove(key.toNSString());
	}

	/** Assign an object to a key. */
	void opIndexAssign(Object object, Object key) {
		return invokeObjcSuper!(void, "setObject:forKey:")(object, key);
	}
	/** ditto */
	final void opIndexAssign(Object object, char[] key) {
		opIndexAssign(object, key.toNSString());
	}

	mixin ObjcBindMethod!(remove, void, "removeObjectForKey:", Object);
	mixin ObjcBindMethod!(opIndexAssign, void, "setObject:forKey:", Object, Object);
	

	// NSExtendedMutableDictionary

	/** Add values from other dictionary to this one. */
	void opCatAssign(NSDictionary other) {
		return invokeObjcSuper!(void, "addEntriesFromDictionary:")(other);
	}

	/** Remove all entries from this dictionary. */
	void removeAll() {
		return invokeObjcSuper!(void, "removeAllObjects")();
	}

	/** Remove all entries where the key is found in keys. */
	void removeAll(NSArray keys) {
		return invokeObjcSuper!(void, "removeObjectsForKeys:")(keys);
	}
	
	/** Replace all the entries in this dictionary with those found in other. */
	void opAssign(NSDictionary other) {
		return invokeObjcSuper!(void, "setDictionary:")(other);
	}
	
	mixin ObjcBindMethod!(opCatAssign, void, "addEntriesFromDictionary:", NSDictionary);
	mixin ObjcBindMethod!(removeAll, void, "removeAllObjects");
	mixin ObjcBindMethod!(removeAll, void, "removeObjectsForKeys:", NSArray);
	mixin ObjcBindMethod!(opAssign, void, "setDictionary:", NSDictionary);


	// NSMutableDictionaryCreation 

	/** Create a dictionary optimized for a specific capacity. */
	this(uint capacity) {
		this(objcClusterAlloc);
		objcInstance = invokeObjcSuper!(id, "initWithCapacity:")(capacity);
	}
	
	// Repeating NSDictionaryCreation

	this() {
		super(objcClusterAlloc);
		objcInstance = invokeObjcSuper!(id, "init")();
	}

	this(NSString path) {
		this(objcClusterAlloc);
		objcInstance = invokeObjcSuper!(id, "initWithContentOfFile:")(path);
	}

//	this(NSURL url) {
//		this(objcClusterAlloc);
//		objcInstance = invokeObjcSuper!(id, "initWithContentOfURL:")(url);
//	}

	this(Object[Object] aa) {
		this(aa.values, aa.keys);
	}

	this(Object[char[]] aa) {
		this(aa.values.toObjcInstances(), aa.keys.toObjcInstances());
	}

	this(Object[] objects, Object[] keys)
	in { assert(objects.length == keys.length); }
	body {
		this(objects.toObjcInstances(), keys.toObjcInstances());
	}

	this(objc.runtime.id[] objects, objc.runtime.id[] keys) 
	in { assert(objects.length == keys.length); }
	body {
		this(objcClusterAlloc);
		objcInstance = invokeObjcSuper!(id, "initWithObjects:forKeys:count:")(objects.ptr, keys.ptr, objects.length);
	}
	
	this(NSArray objects, NSArray keys) {
		this(objcClusterAlloc);
		objcInstance = invokeObjcSuper!(id, "initWithObjects:forKeys:")(objects, keys);
	}
	
	this(NSDictionary otherDictionary) {
		this(objcClusterAlloc);
		objcInstance = invokeObjcSuper!(id, "initWithDictionary:")(otherDictionary);
	}

	this(NSDictionary otherDictionary, bool copyItems) {
		this(objcClusterAlloc);
		objcInstance = invokeObjcSuper!(id, "initWithDictionary:copyItems:")(otherDictionary, copyItems);
	}

}



import cocoa.foundation.autoreleasepool;

unittest {
	scope NSAutoreleasePool pool = new NSAutoreleasePool;
	NSMutableDictionary dict = new NSMutableDictionary;
	
	assert(dict["z"] is null );
	
	dict["key1"] = "obj1".toNSString();
	assert(dict["key1"] == "obj1".toNSString());
	
	dict["key2"] = "obj2".toNSString();
	assert(dict["key2"] == "obj2".toNSString());
	
	dict["key3"] = "obj3".toNSString();
	assert(("key3" in dict) == "obj3".toNSString());
	
	dict.remove("key3");
	assert(("key3" in dict) is null);

	dict["key5"] = "obj5".toNSString();
	dict["key5"] = "obj5 #2".toNSString();
	assert(dict["key5"] == "obj5 #2".toNSString());
	
	char[][char[]] content;
	foreach (key, value; dict) {
		content[key.toString] = value.toString;
	}
	assert(("key1" in content) && content["key1"] == "obj1");
	assert(("key2" in content) && content["key2"] == "obj2");
	assert(("key5" in content) && content["key5"] == "obj5 #2");
	assert(content.length == 3);
}
