
// The D/Objective-C Bridge
// Copyright © 2007  Michel Fortin
// <http://michelf.com/d-objc-bridge/>
//
// This program is free software: you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free 
// Software Foundation, either version 3 of the License, or (at your option) 
// any later version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT 
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for 
// more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program. If not, see <http://www.gnu.org/licenses/>.

module cocoa.foundation.object;

import cocoa.foundation.range;
import cocoa.foundation.string;

import objc.bridge;
import objc.runtime;
import objc.method;
import objc.wrapper : ObjcWrapperBase;

import std.string;
import object;


class NSObject : ObjcWrapperBase {
	mixin ObjcWrap;
	this(objc.runtime.id instance) { super(instance); }

	/// Construct object.
	this() {
		this(objcAlloc);
		objcInstance = invokeObjcSuper!(id, "init")();
	}


	/** Check for object equality with another Object. [self isEqual:object] */
	override int opEquals(Object object) {
		return invokeObjcSuper!(BOOL, "isEqual:")(object);
	}
	
	/** Calculate a hash for this object, calling. [self hash] */
	override uint toHash() {
		return invokeObjcSuper!(uint, "hash")();
	}
	
	/** Get the description of this object. [self description].toString */
	override char[] toString() {
		return invokeObjcSuper!(NSString, "description")().toString;
	}
	
	
	// Special receiver method for opEquals/isEqual: and toString/description
	// to avoid return type mismatch.
	
	private static BOOL objcForward_isEqual_(id self, SEL sel, id object) {
		return objc.exceptionbridge.callD( 
			objcWrapInstance(self).opEquals(objc.type.decapsulate!(Object)(object))
			);
	}
	private static id objcForward_description(id self, SEL sel) {
		return objc.exceptionbridge.callD( 
			objcWrapInstance(self).toString().toNSString().objcInstance
			);
	}
	mixin ObjcMethodInfo!(objcForward_isEqual_, BOOL, "isEqual:", id);
	mixin ObjcMethodInfo!(objcForward_description, id, "description");

	mixin ObjcBindMethod!(toHash, uint, "hash");
	mixin ObjcBindMethod!(toHash, uint, "description");
	
	
	/** Tell if this object is a proxy. */
	bool isProxy() {
		return invokeObjcSuper!(bool, "isProxy")();
	}
	
	/** Tell if this ObjC object is of specified class. */
	bool isKindOfClass(Class c) {
		return invokeObjcSuper!(bool, "isKindOfClass:")(c);
	}
	
	/** Tell if this ObjC object is of specified class or a subclass of it. */
	bool isMemberOfClass(Class c) {
		return invokeObjcSuper!(bool, "isMemberOfClass:")(c);
	}
	
	bool conformsToProtocol(Protocol p) {
		return invokeObjcSuper!(bool, "conformsToProtocol:")(p);
	}

	mixin ObjcBindMethod!(isProxy, bool, "isProxy");
	mixin ObjcBindMethod!(isKindOfClass, bool, "isKindOfClass:", Class);
	mixin ObjcBindMethod!(isMemberOfClass, bool, "isMemberOfClass:", Class);
	mixin ObjcBindMethod!(conformsToProtocol, bool, "conformsToProtocol:", Protocol);


	bool respondsToSelector(SEL selector) {
		return invokeObjcSuper!(bool, "respondsToSelector:")(selector);
	}
	
	mixin ObjcBindMethod!(respondsToSelector, bool, "respondsToSelector:", SEL);
	
	
// Not sure yet what to do about these:
	
//	Object copy() {
//	}
//	
//	Object copyWithZone(NSZone* zone) {
//	}
//	
//	Object mutableCopy() {
//	}
//	
//	Object mutableCopyWithZone(NSZone* zone) {
//	}
//	
//	mixin ObjcBindMethod!(idup, Object, "copy");
//	mixin ObjcBindMethod!(idup, Object, "copyWithZone:");
//	mixin ObjcBindMethod!(dup, Object, "mutableCopy");
//	mixin ObjcBindMethod!(dup, Object, "mutableCopyWithZone:");

	
	// -- Object retention
	// This will be moved to objc.wrapper.ObjcWrapperBase in due time.
	
	invariant {
		// Avoiding a call to the retainCount() method to avoid
		// invariant being checked recursivly.
		uint retainCount = invokeObjcSuper!(uint, "retainCount")();
		assert(retainCount > 0);
	}

	/**
	 * Hash table to store retained objects. This prevents the garbage 
	 * collector from collecting objects retained on the Objective-C side.
	 */
	private static NSObject[objc.runtime.id] retainedObjects;
	
	/**
	 * Give the current retain count for the object. A retain count of 1
	 * means that the object is retained only by the D wrapper object.
	 */
	final uint retainCount() {
		return invokeObjcSuper!(uint, "retainCount")();
	}
	
	/**
	 * Increase the retain count for the object. If retain count grows over
	 * one, object is added to list of retained objects.
	 */
	private final objc.runtime.id retain() {
		if (retainCount == 1) retainedObjects[context.receiver] = this;
		return invokeObjcSuper!(objc.runtime.id, "retain")();
	}
	
	/**
	 * Decrease the retain count for the object. If retain count becomes only
	 * one, object is removed to list of retained objects.
	 */
	private final void release() {
		invokeObjcSuper!(void, "release")();
		if (retainCount == 1) retainedObjects.remove(context.receiver);
	}
	
	mixin ObjcBindMethod!(retain, objc.runtime.id, "retain");
	mixin ObjcBindMethod!(release, void, "release");
	
	/**
	 * Object destructor release the 
	 */
	~this() {
		debug if (objcHardWired) {
			assert(retainCount == 1, 
				"Destructing hard-wired object "~ this.classinfo.name ~" "
				"which is still retained on the Objective-C side."
				);
		}
		// Call directly to avoid rechecking invariants.
		msgSendSuper!(void)(&context, selector!("release"));
	}

}


unittest {
	auto object1 = new NSObject;
	auto object2 = new NSObject;
	
	assert(object1 == object1);
	assert(object1 != object2);
	
	// Check that Objective-C methods are properly mapped by calling
	// the equivalent of [test isEqual:object1], which should return true 
	// only if TestObject.opEquals gets called.
	class TestObject : NSObject {
		override int opEquals(Object o) { return true; }
	}
	TestObject test = new TestObject;
	assert(test.invokeObjcSelf!(BOOL, "isEqual:")(new Object));
}


// Initializing the Objective-C runtime using NSObject as root wrapper object.
static this() {
	objc.wrapper.defaultWrapperFunc = &NSObject.objcCreateWrapper;
}
