
// The D/Objective-C Bridge
// Copyright © 2007  Michel Fortin
// <http://michelf.com/d-objc-bridge/>
//
// This program is free software: you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free 
// Software Foundation, either version 3 of the License, or (at your option) 
// any later version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT 
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for 
// more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program. If not, see <http://www.gnu.org/licenses/>.

module cocoa.foundation.string;

import cocoa.foundation.runtime;
import cocoa.foundation.object;
import cocoa.foundation.range;
import cocoa.foundation.dictionary;
import cocoa.foundation.array;

import objc.bridge;
import objc.runtime;

import std.string;
import std.utf;


/**
 * Create a NSString from a char array (UTF-8 characters string).
 */
NSString toNSString(char[] string) {
	return new D_String!(char)(string);
}

/**
 * Create a NSString from a wchar array (UTF-16 characters string).
 */
NSString toNSStringw(wchar[] string) {
	return new D_String!(wchar)(string);
}

/**
 * Create a NSString from a dchar array (UTF-32 characters string).
 */
NSString toNSStringd(dchar[] string) {
	return new D_String!(dchar)(string);
}


/**
 * Wrapper for Objective-C class NSString.
 *
 * Subclassing notes: NSString is a cluster class. If you subclass it, you need
 * to create your own Objective-C instance first with objcAlloc and pass it as 
 * the argument for the NSString constructor. Otherwise, if you call another
 * NSString constructor which creates the Objective-C instance for you, that
 * instance will not be a capsule for your subclass.
 */
class NSString : NSObject {
	mixin ObjcClusterWrap;
	this(id instance) { super(instance); }
	
	enum Encoding {
		ASCII = 1,		/* 0..127 only */
		NEXTSTEP = 2,
		JapaneseEUC = 3,
		UTF8 = 4,
		ISOLatin1 = 5,
		Symbol = 6,
		NonLossyASCII = 7,
		ShiftJIS = 8,
		ISOLatin2 = 9,
		Unicode = 10,
		WindowsCP1251 = 11,    /* Cyrillic; same as AdobeStandardCyrillic */
		WindowsCP1252 = 12,    /* WinLatin1 */
		WindowsCP1253 = 13,    /* Greek */
		WindowsCP1254 = 14,    /* Turkish */
		WindowsCP1250 = 15,    /* WinLatin2 */
		ISO2022JP = 21,         /* ISO 2022 Japanese encoding for e-mail */
		MacOSRoman = 30,

		Proprietary = 65536,    /* Installation-specific encoding */
	}
	
	
	uint length() {
		return invokeObjcSuper!(uint, "length")();
	}
	
	wchar opIndex(uint index) {
		return invokeObjcSuper!(wchar, "characterAtIndex:")(index);
	}
	
	mixin ObjcBindMethod!(length, uint, "length");
	mixin ObjcBindMethod!(opIndex, wchar, "characterAtIndex:", uint);


	// NSStringExtensionMethods
	
	void getCharacters(wchar* buffer) {
		return invokeObjcSuper!(void, "getCharacters:")(buffer);
	}
	
	void getCharacters(wchar* buffer, NSRange range) {
		return invokeObjcSuper!(void, "getCharacters:range:")(buffer, range);
	}
	
	mixin ObjcBindMethod!(getCharacters, void, "getCharacters:", wchar*);
	mixin ObjcBindMethod!(getCharacters, void, "getCharacters:range:", wchar*, NSRange);
	
	
	NSString opSlice() {
		return this;
	}
	
	NSString opSlice(uint from, uint to) {
		return invokeObjcSuper!(NSString, "substringWithRange:")(NSRange[from..to]);
	}
	
	// Special forwarder function for opSlice converting NSRange in a from..to form.
	private static objc.runtime.id objcForward_substringWithRange_(id self, SEL sel, NSRange range) {
		return objc.exceptionbridge.callD( 
			objcWrapInstance(self).opSlice(range.location, range.end).objcInstance
			);
	}
	mixin ObjcMethodInfo!(objcForward_substringWithRange_, NSString, "subarrayWithRange:", NSRange);
	
	
	NSComparisonResult compare(NSString string) {
		return invokeObjcSuper!(NSComparisonResult, "compare:")(string);
	}
	
	NSComparisonResult compare(NSString string, uint options) {
		return invokeObjcSuper!(NSComparisonResult, "compare:options:")(string, options);
	}
	
	NSComparisonResult compare(NSString string, uint options, NSRange range) {
		return invokeObjcSuper!(NSComparisonResult, "compare:options:range:")(string, options, range);
	}
	
	NSComparisonResult compare(NSString string, uint options, NSRange range, NSDictionary locale) {
		return invokeObjcSuper!(NSComparisonResult, "compare:options:range:locale:")(string, options, range, locale);
	}
	
	mixin ObjcBindMethod!(compare, NSComparisonResult, "compare:", NSString);
	mixin ObjcBindMethod!(compare, NSComparisonResult, "compare:options:", NSString, uint);
	mixin ObjcBindMethod!(compare, NSComparisonResult, "compare:options:range:", NSString, uint, NSRange);
	mixin ObjcBindMethod!(compare, NSComparisonResult, "compare:options:range:locale:", NSString, uint, NSRange, NSDictionary);


	// D comparaison operator.
	
	int opCmp(char[] string) {
		return isEqualToString(string.toNSString());
	}

	
	NSComparisonResult caseInsensitiveCompare(NSString string) {
		return invokeObjcSuper!(NSComparisonResult, "caseInsensitiveCompare:")(string);
	}
	
	NSComparisonResult localizedCompare(NSString string) {
		return invokeObjcSuper!(NSComparisonResult, "localizedCompare:")(string);
	}
	
	NSComparisonResult localizedCaseInsensitiveCompare(NSString string) {
		return invokeObjcSuper!(NSComparisonResult, "localizedCaseInsensitiveCompare:")(string);
	}
	
	mixin ObjcBindMethod!(caseInsensitiveCompare, NSComparisonResult, "caseInsensitiveCompare:", NSString);
	mixin ObjcBindMethod!(localizedCompare, NSComparisonResult, "localizedCompare:", NSString);
	mixin ObjcBindMethod!(localizedCaseInsensitiveCompare, NSComparisonResult, "localizedCaseInsensitiveCompare:", NSString);
	
	
	bool isEqualToString(NSString string) {
		return invokeObjcSuper!(bool, "isEqualToString:")(string);
	}
	
	int opEquals(char[] string) {
		return isEqualToString(string.toNSString());
	}
	
	alias NSObject.opEquals opEquals;
	
	mixin ObjcBindMethod!(isEqualToString, bool, "isEqualToString:", NSString);


	bool hasPrefix(NSString string) {
		return invokeObjcSuper!(bool, "hasPrefix:")(string);
	}

	bool hasSuffix(NSString string) {
		return invokeObjcSuper!(bool, "hasSuffix:")(string);
	}

	mixin ObjcBindMethod!(hasPrefix, bool, "hasPrefix:", NSString);
	mixin ObjcBindMethod!(hasSuffix, bool, "hasSuffix:", NSString);


/* These methods return length==0 if the target string is not found. So, to check for containment: ([str rangeOfString:@"target"].length > 0).  Note that the length of the range returned by these methods might be different than the length of the target string, due composed characters and such.
*/

	NSRange rangeOfString(NSString string) {
		return invokeObjcSuper!(NSRange, "rangeOfString:")(string);
	}
	
	NSRange rangeOfString(NSString string, uint options) {
		return invokeObjcSuper!(NSRange, "rangeOfString:options:")(string, options);
	}
	
	NSRange rangeOfString(NSString string, uint options, NSRange range) {
		return invokeObjcSuper!(NSRange, "rangeOfString:options:range:")(string, options, range);
	}
	
	mixin ObjcBindMethod!(rangeOfString, NSRange, "rangeOfString:", NSString);
	mixin ObjcBindMethod!(rangeOfString, NSRange, "rangeOfString:options:", NSString, uint);
	mixin ObjcBindMethod!(rangeOfString, NSRange, "rangeOfString:options:range:", NSString, uint, NSRange);


	NSString opCat(NSString string) {
		return invokeObjcSuper!(NSString, "stringByAppendingString:")(string);
	}

	final NSString opCat(char[] string) {
		return opCat(string.toNSString());
	}

	mixin ObjcBindMethod!(opCat, NSString, "stringByAppendingString:", NSString);


	double toDouble() {
		return invokeObjcSuper!(double, "doubleValue")();
	}

	float toFloat() {
		return invokeObjcSuper!(float, "floatValue")();
	}

	int toInt() {
		return invokeObjcSuper!(int, "intValue")();
	}

	mixin ObjcBindMethod!(toDouble, double, "doubleValue");
	mixin ObjcBindMethod!(toFloat, float, "floatValue");
	mixin ObjcBindMethod!(toInt, int, "intValue");
	

	NSString toUppercase() {
		return invokeObjcSuper!(NSString, "uppercaseString")();
	}

	NSString toLowercase() {
		return invokeObjcSuper!(NSString, "uppercaseString")();
	}

	NSString toCapitalized() {
		return invokeObjcSuper!(NSString, "uppercaseString")();
	}

	mixin ObjcBindMethod!(toUppercase, NSString, "uppercaseString");
	mixin ObjcBindMethod!(toLowercase, NSString, "lowercaseString");
	mixin ObjcBindMethod!(toCapitalized, NSString, "capitalizedString");
	
	
	char* toUTF8String() {
		return invokeObjcSuper!(char*, "UTF8String")();
	}
	
	mixin ObjcBindMethod!(toUTF8String, char*, "UTF8String");


	// Creation methods

	this() {
		this(objcClusterAlloc);
		objcInstance = invokeObjcSuper!(id, "init")();
	}

	this(char[] characters) {
		this(cast(byte[])characters, Encoding.UTF8);
	}
	
	this(byte[] bytes, uint encoding) {
		this(objcClusterAlloc);
		objcInstance = invokeObjcSuper!(id, "initWithBytes:length:encoding:")(bytes.ptr, bytes.length, encoding);
	}


	version (MAC_OS_X_VERSION_10_4) {

		mixin ObjcMethodWrap!("writeToURL", bool, "writeToURL:atomically:encoding:error:", NSURL, bool, NSStringEncoding, NSError);

		mixin ObjcMethodWrap!("writeToFile", bool, "writeToFile:atomically:encoding:error:", NSString, bool, NSStringEncoding, NSError);

	}

	
	// D UTF-8 string conversion

	char[] toString() {
		return toUTF8String[0..this.length];
	}

	
}


unittest {
	// Check for bugs when two D objects targets the same ObjC instance.
	scope a = new NSString("");
	scope b = new NSString("");
}


/**
 * Wrapper for Objective-C class NSMutableString.
 *
 * Subclassing notes: NSMutableString is a cluster class. If you subclass it, 
 * you need to create your own Objective-C instance first with objcAlloc and 
 * pass it as the argument for the NSMutableString constructor. Otherwise, if 
 * you call another NSMutableString constructor which creates the Objective-C 
 * instance for you, that instance will not be a capsule for your subclass.
 */
class NSMutableString : NSString {
	mixin ObjcClusterWrap;
	this(id instance) { super(instance); }
	
	void opSliceAssign(NSString replacement, uint from, uint to) {
		invokeObjcSuper!(void, "replaceCharactersInRange:withString:")(NSRange[from..to], replacement);
	}
	
	void opSliceAssign(char[] replacement, uint from, uint to) {
		opSliceAssign(replacement.toNSString(), from, to);
	}
	
	// Special forwarder function for opSliceAssign converting NSRange in a from..to form.
	private static void objcForward_replaceCharactersInRange_withString_(id self, SEL sel, NSRange range, id string) {
		objc.exceptionbridge.callD( 
			objcWrapInstance(self).opSliceAssign(NSString.objcWrapInstance(string), range.location, range.end)
			);
	}
	mixin ObjcMethodInfo!(objcForward_replaceCharactersInRange_withString_, void, "replaceCharactersInRange:withString:", NSRange, NSString);

	
	// NSMutableStringExtensionMethods
	
	void insert(NSString string, uint index) {
		invokeObjcSuper!(void, "insertString:atIndex:")(string, index);
	}
	
	void remove(NSRange range) {
		invokeObjcSuper!(void, "deleteCharacters:")(range);
	}
	
	void opCatAssign(NSString string) {
		invokeObjcSuper!(void, "appendString:")(string);
	}
	
	final void opCatAssign(char[] string) {
		opCatAssign(string.toNSString());
	}
	
	void opAssign(NSString string) {
		invokeObjcSuper!(void, "setString:")(string);
	}
	
	final void opAssign(char[] string) {
		opAssign(string.toNSString());
	}
	
	mixin ObjcBindMethod!(insert, void, "insertString:atIndex:", NSString, uint);
	mixin ObjcBindMethod!(remove, void, "deleteCharacters:", NSRange);
	mixin ObjcBindMethod!(opCatAssign, void, "appendString:", NSString);
	mixin ObjcBindMethod!(opAssign, void, "setString:", NSString);
		

	// NSMutableString Creation
	
	this(uint capacity) {
		this(objcClusterAlloc);
		objcInstance = invokeObjcSuper!(id, "initWithCapacity:")(capacity);
	}


	// NSString Creation methods

	this() {
		this(objcClusterAlloc);
		objcInstance = invokeObjcSuper!(id, "init")();
	}

	this(char[] characters) {
		this(cast(byte[])characters, Encoding.UTF8);
	}
	
	this(byte[] bytes, uint encoding) {
		this(objcClusterAlloc);
		objcInstance = invokeObjcSuper!(id, "initWithBytes:length:encoding:")(bytes.ptr, bytes.length, encoding);
	}
	
}


unittest {
	scope auto a = "".toNSString();
}


/**
 * Options for the various search/find and comparison methods (except where noted).
 */
enum {
    NSCaseInsensitiveSearch = 1,
	
	/** Exact character-by-character equivalence. */
    NSLiteralSearch = 2,
	
	/** Search from end of source string. */
    NSBackwardsSearch = 4,
	
	/** Search is limited to start (or end, if NSBackwardsSearch) of source string */
    NSAnchoredSearch = 8, 
	
	/** Added in 10.2; Numbers within strings are compared using numeric value, that is, Foo2.txt < Foo7.txt < Foo25.txt; only applies to compare methods, not find */
    NSNumericSearch = 64
}


private class D_String(C) : NSString {
	static assert(is(C == char) || is(C == wchar) || is(C == dchar));	

	final C[] storage;
	
	this() {
		super(objcAlloc);
		objcInstance = invokeObjcSuper!(id, "init")();
	}
	
	this(C[] characters)
	in {
		validate(characters);
	}
	body {
		this();
		storage = characters.dup;
	}
	
	
	wchar characterAtIndex(uint index) {
		return storage[index];
	}
	
	/**
	 * Return the length of the string.
	 *
	 * This is actually the number of wchar for an UTF-16 string; surrogate 
	 * pairs are considered two characters.
	 */
	uint length() {
		static if (is(C == wchar)) {
			return storage.length;
		} else {
			// TODO: Optimize to avoid creating a new string.
			return storage.toUTF16().length;
		}
	}


	// (NSStringExtensionMethods)
	
	void getCharacters(wchar* buffer) {
		buffer[0..storage.length] = storage.toUTF16();
	}
		
	void getCharacters(wchar* buffer, NSRange range) {
		// TODO: Optimize that: no need to convert the whole string to UTF-16.
		buffer[0..range.length] = storage.toUTF16()[range.location..range.end];
	}
	
	static if (is(C == wchar)) {
	
		NSString substringFromIndex(uint from) {
			return new D_String(storage[from..$]);
		}
		
		NSString substringToIndex(uint to) {
			return new D_String(storage[0..to]);
		}
		
		NSString substringWithRange(NSRange range) {
			return new D_String(storage[range.location..range.end]);
		}
	
	}
	
	alias NSString.compare compare;
	
	static if (is(C == char)) {
		// std.string.cmp only works on char
	
		NSComparisonResult compare(NSString string) {
			if (auto d_string = cast(D_String)string) {
				int result = std.string.cmp(this.storage, d_string.storage);
				return cast(NSComparisonResult)result;
			}
			return super.compare(string);
		}
		
		// Buggy: range is expressed in term of UTF-16 length, not UTF-8 as this function implies.
//		NSComparisonResult compare(NSString string, uint options, NSRange compareRange) {
//			if (options == NSLiteralSearch) {
//				if (auto d_string = cast(D_String)string) {
//					with (compareRange) {
//						// Slice...
//						char[] first  = this.storage[start..end];
//						char[] second = d_string.storage[start..end];
//						// ... and compare
//						int result = std.string.cmp(first, second);
//						return cast(NSComparisonResult)result;
//					}
//				}
//			}
//			return super.compare(string, options, compareRange);
//		}
	
	}
	
	override bool isEqualToString(NSString string) {
		if (auto d_string = cast(D_String)string) {
			return this.storage == d_string.storage;
		} else {
			return super.isEqualToString(string);
		}
	}

	static if (is(C == char)) {
	
		override NSString opCat(NSString string) {
			return new D_String(storage ~ string.toString);
		}
		
		alias NSString.opCat opCat;
	
	} else {
	
		override NSString opCat(NSString string) {
			if (auto d_string = cast(D_String)string) {
				return opCat(d_string);
			}
			return super.opCat(string);
		}
		
		alias NSString.opCat opCat;
	
	}
	
	char* utf8String() {
		return storage.toUTF8().toStringz();
	}

	// D character array behaviour
	
	static if (is(C == wchar)) {

		NSString opSlice(uint from, uint to) {
			return new D_String(storage[from..to]);
		}
		
		alias NSString.opSlice opSlice;
		
	}
	
	static if (is(C == char)) {
		
		int opCompare(C[] string) {
			return std.string.cmp(this.storage, string);
		}

		D_String opCat(C[] string) {
			return new D_String(this.storage ~ string);
		}
	
	}

	char[] toString() {
		return storage.toUTF8();
	}
	
}


import cocoa.foundation.autoreleasepool;

unittest {
	scope NSAutoreleasePool pool = new NSAutoreleasePool;

	scope a = new NSString("string");
	scope b = new D_String!(char)("string");
	scope c = "string".toNSString();
	
	assert(a.isEqualToString(b));
	assert(a == b);
	assert(b == c);
	
	// Check operations with multibyte UTF-8 sequences.
	assert("Français".toNSString().length == 8);
	assert("русский язык".toNSString().length == 12);
	assert((a ~ "日本語").length == a.length + 3);

}


