
// The D/Objective-C Bridge
// Copyright © 2007  Michel Fortin
// <http://michelf.com/d-objc-bridge/>
//
// This program is free software: you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free 
// Software Foundation, either version 3 of the License, or (at your option) 
// any later version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT 
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for 
// more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program. If not, see <http://www.gnu.org/licenses/>.

module objc.bridge;

static import objc.subclassinit;
static import objc.wrapper;

public import objc.type : ObjcObject;


// -- Deriving Classes

/**
 * Shortcut template that takes directly the name of the enclosing class and
 * use it as the Objective-C subclass name.
 */
template ObjcSubclass() {
	mixin ObjcSubclass!(this.stringof);
}


/**
 * Template to pair a new Objective-C subclass to a derivative of ObjcObject.
 */
template ObjcSubclass(char[] className) {

	static assert(is(typeof(this) : objc.type.ObjcObject), 
		typeof(this).stringof ~" has Objective-C bindings but does "
		"not declare the ObjcObject interface.");

	// Required mixins.
	mixin objc.subclassinit.ObjcSubclassInitializer!(className);
	
	static if (is(typeof(this) : objc.wrapper.ObjcWrapperBase)) {
		/**
		 * Tell if the linked Objective-C instance is hard-wired to this specific
		 * D object and cannot exist without it. Only available in debug mode.
		 */
		debug protected override bool objcHardWired() {
			return true;
		}
	}
	
	private alias typeof(this) HostObject;
	
	static HostObject objcWrapInstance(objc.runtime.id instance) {
		static if (is(typeof(this) : objc.wrapper.ObjcWrapperBase)) {
			return cast(HostObject)objc.objectlookup.lookup(instance);
		} else {
			return cast(HostObject)objc.capsule.removeCapsule(instance);
		}
	}
	
	// Create capsule instance.
	static if (!is(typeof(&(new typeof(this)).objcInstance).stringof == objc.runtime.id delegate())) {
		objc.runtime.id objcInstance() {
			return objc.capsule.createCapsule(this, objcClass);
		}
	}
	
}


/**
 * Shortcut template that takes directly the name of the enclosing class and
 * use it as the Objective-C subclass name.
 */
template ObjcWrap() {
	mixin ObjcWrap!(this.stringof);
}



/**
 * Template to pair an existing Objective-C class to a derivative of 
 * ObjcWrapperObject.
 */
template ObjcWrap(char[] className) {
	
	static assert(is(typeof(this) : objc.wrapper.ObjcWrapperBase), 
		typeof(this).stringof ~" wants to be an Objective-C wrapper but is "
		"not a subclass of ObjcWrapperBase.");

	// Required mixins.
	mixin objc.subclassinit.ObjcSubclassInitializer!("D_" ~ className, className);
	mixin objc.wrapper.ObjcWrapperDerived;
	
	/**
	 * Tell if the linked Objective-C instance is hard-wired to this specific
	 * D object and cannot exist without it. Only available in debug mode.
	 */
	debug protected bool objcHardWired() {
		return context.receiver.isa == objcClass;
	}
	
	mixin ObjcObjectWrapBase;

}


/**
 * Shortcut template that takes directly the name of the enclosing class and
 * use it as the Objective-C subclass name.
 */
template ObjcClusterWrap() {
	mixin ObjcClusterWrap!(this.stringof);
}


/**
 * Template to pair an existing Objective-C class cluster to a derivative of 
 * ObjcWrapperObject.
 */
template ObjcClusterWrap(char[] className) {

	mixin ObjcWrap!(className);
	
	/**
	 * Special Objective-C instance allocator function adapted to 
	 * cluster class wrappers. Invoke "alloc" on superclass instead 
	 * of this class.
	 */
	static objc.runtime.id objcClusterAlloc() {
		return invokeObjcSuperClass!(objc.runtime.id, "alloc")();
	}

}



/**
 * Template to mixin within Objective-C binding templates. Will create a
 * capsule Objective-C subclass automatically for the enclosing class if
 * none is already defined.
 */
template ObjcSubclassIfNeeded() {
	static if (
		// Checking if a subclass exists...
		!is(typeof(typeof(this).objcClassDef) == objc.runtime.objc_class) ||
		
		// ...and if the subclass is specific to this (not comming from super).
		(is(typeof(typeof(super).objcClassDef) == objc.runtime.objc_class) 
			&& typeof(this).objcClassName == typeof(super).objcClassName)
		)
	{
		// If so, add a new subclass to this class.
		mixin ObjcSubclass;
	}
}


/**
 * Template which can be mixed in a class to bind a selector sel to a
 * constructor.
 *
 * This will create a receiver function which will forward the call to
 * the constructor matching specified arguments, decapsulating arguments 
 * when appropriate.
 */
template ObjcBindInitializer(char[] sel, A...) {
	mixin ObjcSubclassIfNeeded;
	
	alias objc.method.ObjcMethodSelectorCheck!(sel, A) check;
	
	private alias typeof(this) HostObject;
	
	/**
	 * Call forwarder which can be called non-virtualy to forward its
	 * call to a virtual method.
	 */
	private static objc.runtime.id objcInit(objc.runtime.id self, objc.runtime.SEL sel, A args) {
		// When calling this function, self must be a statically-allocated 
		// placeholder (fake) object, so we must not release it.
		assert(self == objcPlaceholder, 
			"Unable to initialize: self is not the placeholder object for "
			"class "~ objcClassName ~".");
			
		objc.runtime.id construct() {
			// Automatically convert arguments.
			std.typetuple.TypeTuple!(A) args;
			foreach (i, Type; A) {
				args[i] = objc.type.decapsulate!(Type)(objcArgs[i]);
			}
		
			// Call constrctor and return Objective-C instance.
			auto obj = new HostObject(args);
			return obj.objcInstance;
		}
		
		return objc.exceptionbridge.callD( construct() );
	}
	
	mixin ObjcInitMethodInfo!(objcInit, objc.runtime.id, sel, A);
}


/**
 * Template which can be mixed in a class to bind selector sel to instance 
 * method.
 *
 * This will create a receiver function which will forward the call to
 * method, decapsulating arguments and encapsulating the return value as
 * appropriate.
 */
template ObjcBindMethod(alias method, R, char[] sel, A...) {
	mixin ObjcSubclassIfNeeded;

	alias objc.method.ObjcMethodSelectorCheck!(sel, A) check;
	
	/**
	 * Call forwarder which can be called non-virtualy to forward its
	 * call to a virtual method.
	 */
	private final R msgForwardCall(A args) {
		return method(args);
	}
	
	private alias objc.method.msgReceive!(R, objcWrapInstance, msgForwardCall, A) msgReceive;
	
	mixin ObjcMethodInfo!(msgReceive, R, sel, A);
}


/**
 * Template which can be mixed in a class to bind selector sel to a
 * class (static) method.
 *
 * This will create a receiver function which will forward the call to
 * method, decapsulating arguments and encapsulating the return value as
 * appropriate.
 */
template ObjcBindClassMethod(alias method, R, char[] sel, A...) {
	mixin ObjcSubclassIfNeeded;

	alias objc.method.ObjcMethodSelectorCheck!(sel, A) check;
	
	private alias objc.method.msgReceive!(R, objcWrapInstance, method, A) msgReceive;
	
	/**
	 * Call forwarder which can only call static functions.
	 */
	private static final R msgForwardStaticCall(A args) {
		return method(args);
	}
	
	mixin ObjcClassMethodInfo!(msgForwardStaticCall, R, sel, A);
}


/**
 * Template to mixin within a class to make the subclass initializer aware 
 * of an Objective-C method to define.
 */
template ObjcMethodInfo(alias receiver, R, char[] sel, A...) {
	mixin ObjcSubclassIfNeeded;

	private alias objc.type.ObjcType2!(R) ObjcR;
	private alias objc.type.ObjcTypeTuple!(objc.runtime.id, objc.runtime.SEL, A) ObjcA;

	// This checks that receiver is a function of the right type.
	private static const ObjcR function(ObjcA) typeCheck = &receiver;
	
	private static const char[] methodType = 
		objc.typecoding.methodTypeEncode!(ObjcR, ObjcA);
			
	objc.methodlist.MemberObjcMethod!(sel, methodType, receiver, objc.methodlist.MethodKind.INSTANCE) methodInfo;
	
}


/**
 * Template to mixin within a class to make the subclass initializer aware 
 * of an Objective-C method to define.
 */
template ObjcInitMethodInfo(alias receiver, R, char[] sel, A...) {
	mixin ObjcSubclassIfNeeded;

	private alias objc.type.ObjcType2!(R) ObjcR;
	private alias objc.type.ObjcTypeTuple!(objc.runtime.id, objc.runtime.SEL, A) ObjcA;

	// This checks that receiver is a function of the right type.
	private static const ObjcR function(ObjcA) typeCheck = &receiver;
	
	private static const char[] methodType = 
		objc.typecoding.methodTypeEncode!(ObjcR, ObjcA);
			
	objc.methodlist.MemberObjcMethod!(sel, methodType, receiver, objc.methodlist.MethodKind.INSTANCE_INIT) methodInfo;
	
}


/**
 * Template to mixin within a class to make the subclass initializer aware 
 * of an Objective-C method to define.
 */
template ObjcClassMethodInfo(alias receiver, R, char[] sel, A...) {
	mixin ObjcSubclassIfNeeded;

	private alias objc.type.ObjcType2!(R) ObjcR;
	private alias objc.type.ObjcTypeTuple!(A) ObjcA;

	// This checks that receiver is a function of the right type.
	private static const ObjcR function(ObjcA) typeCheck = receiver;
	
	private static const char[] methodType = 
		objc.typecoding.methodTypeEncode!(ObjcR, 
			objc.runtime.id, objc.runtime.SEL, ObjcA);
			
	objc.methodlist.MemberObjcClassMethod!(sel, methodType, receiver, objc.methodlist.MethodKind.CLASS) methodInfo;
	
}


/**
 * Object wrapper base template to mixin within ObjcObject derivatives when
 * creating a subclass of an Objective-C object.
 */
template ObjcObjectWrapBase() {

	/** Type of the template host object. */
	private alias typeof(this) HostObject;
	
	/** Host needs to be an Object. */
	static assert(is(HostObject : Object));
	
	
	// -- Type wraping

	/**
	 * Return associated D object for ObjC object instance. The ObjC instance
	 * is encapsulated if necessary, and null is returned on classes mismatch.
	 */
	static HostObject objcWrapInstance(objc.runtime.id instance) {
		// Call the initializer, if it exists, when passing here the first time.
		static if (is(typeof(&HostObject.objcWrapInit) == void function())) {
			static bool firstPass = true;
			if (firstPass) {
				firstPass = false;
				HostObject.objcWrapInit();
			}
		}

		Object base = objc.wrapper.wrapInstance(instance);
		if (base is null)  return null;
		
		// Check if wrapInstance returned a compatible object from lookup.
		// If not, we need to create a new wrapper with our derived class.
		HostObject casted = cast(HostObject)base;
		if (casted)  return casted;
		
		casted = cast(HostObject)objc.wrapper.wrapInstanceNoLookup(instance);
		// Note that at this point wrapInstanceNoLookup should be guarentied
		// to return an object of class HostObject (or a derivative class) 
		// since we've registered our callback with objcWrapInit above.
		assert(casted !is null);
		// If the above assertion fails, it probably means the class for 
		// HostObject hasn't been initialized. This is generally done by adding
		// static method objcWrapInit to the class.
		return casted;
	}

}


