
// The D/Objective-C Bridge
// Copyright © 2007  Michel Fortin
// <http://michelf.com/d-objc-bridge/>
//
// This program is free software: you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free 
// Software Foundation, either version 3 of the License, or (at your option) 
// any later version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT 
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for 
// more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program. If not, see <http://www.gnu.org/licenses/>.

module objc.method;

import objc.runtime;
import objc.msg;
import objc.type;
import objc.typecoding;
import objc.exceptionbridge;

import std.typetuple;


//debug = ObjcBridgeLog;



//version = ObjcSelectorPreRegister;


version (ObjcSelectorPreRegister) {

	/** Objective-C selector shortcut template. */
	template selector(char[] name) {
		alias ObjcSelector!(name).sel selector;
	}


	/**
	 * Objective-C selector template with initializer.
	 */
	private template ObjcSelector(char[] name) {
		/** Method selector, statically initialized by constructor. */
		const SEL sel;
		static this() {
			sel = sel_registerName(name);
		}
	}
	
} else {

	objc.runtime.SEL selector(char[] name)() {
		return sel_registerName(name.ptr);
	}

}


/**
 * Send Objective-C message to object instance, encapsulating arguments, 
 * decapsulating return value, and bridging exceptions as necessary.
 */
R msgSend(R, A...)(id self, SEL sel, A args) {
	debug (ObjcBridgeLog) printf("Sending   -[%s %s]\n", self.isa.name, sel);
	
	ObjcTypeTuple!(A) objcArgs;
	ObjcType!(R) objcResult;
	foreach (i, Type; A)  objcArgs[i] = encapsulate!(Type)(args[i]);
	objc.exceptionbridge.callObjc(
		objcResult = objc.msg.send!(ObjcType!(R))(self, sel, objcArgs)
		);
	return decapsulate!(R)(objcResult);
}

/** ditto */
void msgSend(R : void, A...)(id self, SEL sel, A args) {
	debug (ObjcBridgeLog) printf("Sending   -[%s %s]\n", self.isa.name, sel);
	
	ObjcTypeTuple!(A) objcArgs;
	foreach (i, Type; A)  objcArgs[i] = encapsulate!(Type)(args[i]);
	objc.exceptionbridge.callObjc(
		objc.msg.send!(ObjcType!(R))(self, sel, objcArgs)
		);
}

/**
 * Send Objective-C message to object instance superclass, encapsulating 
 * arguments, decapsulating return value, and bridging exceptions as necessary.
 */
R msgSendSuper(R, A...)(objc_super* context, SEL sel, A args) {
	debug (ObjcBridgeLog) printf("Sending   -[%s %s] (self is %s)\n", context.super_class.name, sel, context.receiver.isa.name);
	
	ObjcTypeTuple!(A) objcArgs;
	ObjcType!(R) objcResult;
	foreach (i, Type; A)  objcArgs[i] = encapsulate!(Type)(args[i]);
	objc.exceptionbridge.callObjc(
		objcResult = objc.msg.sendSuper!(ObjcType!(R))(context, sel, objcArgs)
		);
	return decapsulate!(R)(objcResult);
}

/** ditto */
void msgSendSuper(R : void, A...)(objc_super* context, SEL sel, A args) {
	debug (ObjcBridgeLog) printf("Sending   -[%s %s] (self is %s)\n", context.super_class.name, sel, context.receiver.isa.name);
	
	ObjcTypeTuple!(A) objcArgs;
	foreach (i, Type; A)  objcArgs[i] = encapsulate!(Type)(args[i]);
	objc.exceptionbridge.callObjc(
		objc.msg.sendSuper!(ObjcType!(R))(context, sel, objcArgs)
		);
}


unittest {
	if (0) {
		// Just compile this unittest to instanciate the templates. 
		// Actual testing would require an object.
		
		SEL sel = selector!("test");
		objc_super context;
	
		// Try msgSend methods.
		msgSend!(bool)(null, sel);
		msgSend!(bool, bool)(null, sel, 1);
		msgSend!(bool, bool, bool)(null, sel, 1, 0);
		msgSend!(bool, bool, bool, bool)(null, sel, 0, 1, 1);
		msgSend!(bool, bool, bool, bool, bool)(null, sel, 0, 1, 0, 0);
		// Void-variant methods.
		msgSend!(void)(null, sel);
		msgSend!(void, bool)(null, sel, 0);
		msgSend!(void, bool, bool)(null, sel, 0, 1);
		msgSend!(void, bool, bool, bool)(null, sel, 0, 0, 0);
		msgSend!(void, bool, bool, bool, bool)(null, sel, 0, 1, 0, 1);
		
		// Try msgSendSuper methods.
		msgSendSuper!(bool)(&context, sel);
		msgSendSuper!(bool, bool)(&context, sel, 1);
		msgSendSuper!(bool, bool, bool)(&context, sel, 1, 0);
		msgSendSuper!(bool, bool, bool, bool)(&context, sel, 0, 1, 1);
		msgSendSuper!(bool, bool, bool, bool, bool)(&context, sel, 0, 1, 0, 0);
		// Void-variant methods.
		msgSendSuper!(void)(&context, sel);
		msgSendSuper!(void, bool)(&context, sel, 0);
		msgSendSuper!(void, bool, bool)(&context, sel, 0, 1);
		msgSendSuper!(void, bool, bool, bool)(&context, sel, 0, 0, 0);
		msgSendSuper!(void, bool, bool, bool, bool)(&context, sel, 0, 1, 0, 1);
	}
}


/**
 * Receive Objective-C message and forward it to the receiver, decapsulating 
 * arguments, encapsulating return value, and bridging exceptions as necessary.
 */
objc.type.ObjcType2!(R) msgReceive(R, alias resolve, alias method, A...)
	(objc.runtime.id self, objc.runtime.SEL sel, objc.type.ObjcTypeTuple!(A) objcArgs)
{
	debug (ObjcBridgeLog) printf("Receiving -[%s %s]\n", self.isa.name, sel);
	
	objc.type.ObjcType!(R) forward() {
		std.typetuple.TypeTuple!(A) args;
		foreach (i, Type; A) {
			args[i] = objc.type.decapsulate!(Type)(objcArgs[i]);
		}
		
		// Create method call as a delegate.
		R delegate(A) methodCall;
		methodCall.ptr = cast(void*)resolve(self);
		methodCall.funcptr = &method;
		
		R result = methodCall(args);
		return objc.type.encapsulate!(R)(result);
	}
	
	return objc.exceptionbridge.callD(forward());
}


/** ditto */
void msgReceive(R : void, alias resolve, alias method, A...)
	(objc.runtime.id self, objc.runtime.SEL sel, objc.type.ObjcTypeTuple!(A) objcArgs)
{
	debug (ObjcBridgeLog) printf("Receiving -[%s %s]\n", self.isa.name, sel);
	
	void forward() {
		std.typetuple.TypeTuple!(A) args;
		foreach (i, Type; A) {
			args[i] = objc.type.decapsulate!(Type)(objcArgs[i]);
		}
		
		// Create method call as a delegate.
		R delegate(A) methodCall;
		methodCall.ptr = cast(void*)resolve(self);
		methodCall.funcptr = &method;
		
		methodCall(args);
	}
	
	objc.exceptionbridge.callD(forward());
}


/**
 * Check mismatched selectors by checking the number of arguments in
 * sel againt the number of argumentsn in A.
 */
template ObjcMethodSelectorCheck(char[] sel, A...) {

	// Check for the right number of arguments.
	static if (A.length != objc.method.methodArgumentCount!(sel)) {
		// Create a message that will show up when there are missing arguments.
		private const loc = __FILE__ ~":"~ std.metastrings.ToString!(__LINE__) ~": ";
		private const call = "-["~ objcClassName ~" "~ sel ~"]";
		private const margs = std.metastrings.ToString!(objc.method.methodArgumentCount!(sel));
		
		pragma(msg, loc~ 
			call ~" requires "~ margs ~ 
			" arguments.");
	}
	static assert(A.length == objc.method.methodArgumentCount!(sel));

}


/**
 * Number of arguments in method name. This simply count the number of 
 * colon character within the string.
 */
template methodArgumentCount(char[] name) {
	static if (name.length == 0) {
		const uint methodArgumentCount = 0;
	} else static if (name[0] == ':') {
		const uint methodArgumentCount = 1 + methodArgumentCount!(name[1..$]);
	} else {
		const uint methodArgumentCount = methodArgumentCount!(name[1..$]);
	}
}

