
// The D/Objective-C Bridge
// Copyright © 2007  Michel Fortin
// <http://michelf.com/d-objc-bridge/>
//
// This program is free software: you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free 
// Software Foundation, either version 3 of the License, or (at your option) 
// any later version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT 
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for 
// more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program. If not, see <http://www.gnu.org/licenses/>.

/**
 * Definitions of the Objective-C runtime data structures, types and functions.
 */
module objc.runtime;

import std.c.stdlib : malloc, free;


/// ObjC boolean value, defined as one (signed) byte
alias byte BOOL;
enum { NO, YES }


/* 
 * Class
 */
alias objc_class* Class;
struct objc_class {

	/// Points to the metaclass of this class. If this class is a metaclass, 
	/// the isa field points to the root metaclass (usually the metaclass for 
	/// NSObject, but also possibly the metaclass for NSProxy or a root class 
	/// of your own—a root class is any class that does not inherit from 
	/// another class). This also means that the isa field for the metaclass 
	/// of the root class points to itself
	objc_class* isa;
	
	/// Points to the objc_class data structure for the superclass of this 
	/// class, or NULL if this is a root class.
	objc_class* super_class;
	
	/// Points to a C string containing the name of the class.
	char* name;
	
	/// An integer indicating the version number of the class, which you can 
	/// modify at runtime using the class_setVersion function.
	int version_ = 0;
	
	/// Contains a set of bit flags used by the Objective-C runtime functions.
	int info;
	
	/// An integer indicating the size of the instance variables used by this
	/// class. This value includes the value of the instance_size field of the
	/// superclass.
	int instance_size;
	
	/// A pointer to an objc_ivar_list data structure describing the instance
	/// variables that are allocated for each instance of this class. This 
	/// pointer may be NULL, in which case this class has no instance 
	/// variables.
	objc_ivar_list* ivars;
	
	/// If the CLS_METHOD_ARRAY flag is set, this field is an array of 
	/// objc_method_list data structures that collectively specify all the 
	/// instance methods that can be sent to objects that are instances of 
	/// this class. If the CLS_METHOD_ARRAY flag is not set, this field is a 
	/// pointer to a single objc_method_list data structure. If this class is 
	/// a metaclass definition, this field specifies the class methods of the 
	/// class.
	objc_method_list** methodLists;
	
	/// A pointer to an objc_cache method cache data structure.
	objc_cache* cache;
	
	/// A pointer to a objc_protocol_list data structure. This is a list of 
	/// the formal protocols this class claims to implement.
	objc_protocol_list* protocols;
}


enum : int {
	CLS_CLASS = 0x1,
	CLS_META = 0x2,
	CLS_INITIALIZED = 0x4,
	CLS_POSING = 0x8,
	CLS_MAPPED = 0x10,
	CLS_FLUSH_CACHE = 0x20,
	CLS_GROW_CACHE = 0x40,
	CLS_NEED_BIND = 0x80,
	CLS_METHOD_ARRAY = 0x100,
// the JavaBridge constructs classes with these markers
	CLS_JAVA_HYBRID = 0x200,
	CLS_JAVA_CLASS = 0x400,
// thread-safe +initialize
	CLS_INITIALIZING = 0x800,
// bundle unloading
	CLS_FROM_BUNDLE = 0x1000,
// C++ ivar support
	CLS_HAS_CXX_STRUCTORS = 0x2000,
// Lazy method list arrays
	CLS_NO_METHOD_ARRAY = 0x4000,
// +load implementation
//	CLS_HAS_LOAD_METHOD = 0x8000,
}


extern (C) {


	/* -- Accessing the Class Version -- */
	
	/// Sets the version number of a class definition
	void class_setVersion(Class theClass, int version_);

	/// Returns the version number of a class definition
	int class_getVersion(Class theClass);
	
	
	/* -- Posing As Another Class -- */
	
	/// Globally override a specified class definition with a different class 
	/// definition
	Class class_poseAs(Class imposter, Class original);

	
	/* -- Obtaining Class Definitions -- */
	
	/// Obtains the list of registered class definitions
	int objc_getClassList(Class* buffer, int bufferLen);

	/// Returns a pointer to the class definition of the specified class
	Class objc_getClass(char* aClassName);
	
	/// Returns a pointer to the class definition of the specified class
	Class objc_lookUpClass(char *aClassName);
	
	/// Returns a pointer to the metaclass definition of the specified class
	Class objc_getMetaClass(char *aClassName);

	/// Sets a custom error-handling callback function called from 
	/// objc_getClass and objc_getMetaClass when the desired class is not 
	/// found
	void objc_setClassHandler(int function(char*) callback);

	
	/* -- Instantiating Classes -- */
	
	/// Creates an instance of a class, allocating memory for the class in 
	/// the default malloc memory zone
	id class_createInstance(Class theClass, uint additionalByteCount);

	/// Creates an instance of a class, allocating memory for the class in a 
	/// specific malloc memory zone
	id class_createInstanceFromZone(Class theClass, uint additionalByteCount, void* zone);


	/* -- Adding Classes -- */

	/// Registers a class definition with the Objective-C runtime
	void objc_addClass(Class myClass);

	/// Handles creation and registration of a class when objc_getClass or 
	/// objc_getMetaClass cannot otherwise find the class
	alias int function(char* className) ClassHandlerCallback;

}


/*
 * Protocol
 */

typedef objc_object Protocol;

/* List of protocols */
struct objc_protocol_list {

	/// A pointer to another objc_protocol_list data structure
	objc_protocol_list *next;
	
	/// The number of protocols in this list
	int count;
	
	/// An array of pointers to objc_class data structures that represent 
	/// protocols
	Protocol*[0] list;

}


/*
 * Selectors
 */

/// Represents a method selector
typedef char* SEL;


extern (C) {
	
	/// Returns the name of the method specified by the selector.
	char* sel_getName(SEL aSelector);
	
	/// Indicates whether a selector is registered.
	BOOL sel_isMapped(SEL aSelector);
	
	/// Registers a method with the Objective-C runtime system, maps the 
	/// method name to a selector, and returns the selector value.
	SEL sel_registerName(char* str);
	
	/// Registers a method name with the Objective-C runtime system.
	SEL sel_getUid(char* str);


	// -- Senging messages
	
	/// Sends a message with a simple return value to an instance of a class.
	id objc_msgSend(id theReceiver, SEL theSelector, ...);
	
	/// Sends a message with a data-structure return value to an instance of 
	/// a class.
	void objc_msgSend_stret(void* stretAddr, id theReceiver, SEL theSelector, ...);

	version (X86) {
		/// Sends a message with a floating point return value to an instance 
		/// of a class. [X86 only]
		double objc_msgSend_fpret(id self, SEL op, ...);
	}


	/// Sends a message with a simple return value to the superclass of an 
	/// instance of a class
	id objc_msgSendSuper(objc_super* superContext, SEL theSelector, ...);

	/// Sends a message with a data-structure return value to the superclass 
	/// of an instance of a class
	void objc_msgSendSuper_stret(void* stretAddr, objc_super* superContext, SEL theSelector, ...);

	
	/* Forwarding Messages */
	
	/// Given an argument list, send a message with a simple return value
	id objc_msgSendv(id self, SEL op, uint arg_size, marg_list arg_frame);

	/// Given an argument list, send a message with a data-structure return 
	/// value
	void objc_msgSendv_stret(void* stretAddr, id self, SEL op, uint arg_size, marg_list arg_frame);

	version (X86) {
		/// Given an argument list, send a message with a floating point 
		/// return value. [X86 only] */
		double objc_msgSendv_fpret(id self, SEL op, uint arg_size, marg_list arg_frame);
	}

}

/*
 * Method
 */

alias objc_method *Method;
struct objc_method {

	/// A pointer of type SEL. Points to the method selector that uniquely 
	/// identifies the name of this method
	SEL method_name;

	/// A pointer to a C string. This string contains the type encodings for 
	/// the method’s argument. See “Type Encodings” for information on valid 
	/// encoding formats
	char *method_types;
	
	/// A pointer to the start of the method implementation. In a class 
	/// definition that represents a formal protocol, this field is NULL
	IMP method_imp;
}

/// A pointer to the start of a method implementation
alias id function(id, SEL, ...) IMP;

/// Contains an array of method definitions
struct objc_method_list {

	/// Reserved for future use
	void* obsolete;

	/// An integer specifying the number of methods in the method list 
	/// array
	int method_count;

	/// An array of objc_method data structures
	objc_method[0] method_list;
	
}


extern (C) {

	/* -- Accessing Methods -- */
	
	/// Returns a pointer to the data structure describing the specified 
	/// instance method
	Method class_getInstanceMethod(Class aClass, SEL aSelector);

	/// Returns a pointer to the data structure describing the specified 
	/// class method
	Method class_getClassMethod(Class aClass, SEL aSelector);

	/// Returns one of the method lists contained in the specified class 
	/// definition
	objc_method_list* class_nextMethodList(Class theClass, void** iterator);

	/// Adds a list of methods to a class definition
	void class_addMethods(Class aClass, objc_method_list* methodList);

	/// Removes a previously added list of methods from a class definition
	void class_removeMethods(Class aClass, objc_method_list* methodList);

	/// Returns the number of arguments accepted by a method
	uint method_getNumberOfArguments(Method method);
	
	/// Returns the total size of the stack frame occupied by a method’s 
	/// arguments
	uint method_getSizeOfArguments(Method method);

	/// Returns information about one of a method’s arguments
	uint method_getArgumentInfo(Method method, int argIndex, char** type, int* offset);

}

/* Argument List
 * These used to be macros, now they are functions but they should work the same. */

typedef void* marg_list;

version (PPC)  const size_t marg_prearg_size = 128;
else           const size_t marg_prearg_size = 0;

/// Allocates a new argument list
void marg_malloc(out marg_list margs, Method method) {
	margs = cast(marg_list)malloc(marg_prearg_size + 
		((7 + method_getSizeOfArguments(method)) & ~7));
}

/// Releases an argument list
void marg_free(in marg_list margs) {
	free(margs);
}

template marg_values(T) {

	private size_t marg_adjustedOffset(size_t offset) {
		return marg_prearg_size + offset;
	}

	/// Return a pointer to an argument in an argument list
	T* marg_getRef(marg_list margs, size_t offset) {
		return cast(T*)(cast(byte*)margs + marg_adjustedOffset(offset));
	}

	/// Return the value of an argument in an argument list
	T marg_getValue(marg_list margs, size_t offset) {
		return *marg_getRef(margs, offset);
	}

	/// Set the value of an argument in an argument list
	void marg_setValue(marg_list margs, size_t offset, T value) {
		*marg_getRef(margs, offset) = value;
	}
	
}



/*
 * Instance
 */

alias objc_object* id;
struct objc_object {

	/// A pointer to the class definition of which this object is an instance
	Class isa;
	
	/* …variable length data containing instance variable values…  */
}


/*
 * Instance variables
 */

/// Specifies the name, type, and location of one instance variable
alias objc_ivar* Ivar;
struct objc_ivar {

	/// A pointer to a C string containing the name of the instance variable
	char* ivar_name;
	
	/// A pointer to a C string containing the type encoding of the variable. 
	/// See “Type Encodings” for valid type encodings for instance variables
	char* ivar_type;
	
	/// An integer indicating the location of this instance variable within 
	/// the memory allocated for an instance of the class containing this 
	/// variable. The offset is from the start of the instance memory to the 
	/// location of this variable
	int ivar_offset;
}

/// Specifies the instance variables of each instance of a class
struct objc_ivar_list {

	/// An integer specifying a the number of elements in the ivar_list 
	/// array
	int ivar_count;
	
	/// An array of objc_ivar data structures. [variable length]
	objc_ivar[0] ivar_list;

}


extern (C) {


	/* -- Accessing Instance Variable Definitions -- */

	/// Obtains information about the instance variables defined for a 
	/// particular class
	Ivar class_getInstanceVariable(Class aClass, char* aVariableName);


	/* -- Accessing Instance Variables -- */
	
	/// Changes the value of an instance variable of a class instance
	Ivar object_setInstanceVariable(id object, char* name, void* value);

	/// Obtains the value of an instance variable of a class instance
	Ivar object_getInstanceVariable(id object, char* name, void** value);

}


/* -- Callbacks ----------------------------------------------------------- */


/*
 * Misc
 */

/// Performance optimization for method calls. Contains pointers to recently 
/// used methods
struct objc_cache {

	/// An integer specifying the total number of allocated cache buckets 
	/// (minus one). During method lookup, the Objective-C runtime uses this 
	/// field to determine the index at which to begin a linear search of the 
	/// buckets array. A pointer to a method’s selector is masked against 
	/// this field using a logical AND operation (index = (mask & selector)). 
	/// This serves as a simple hashing algorithm
	uint mask;
	
	/// An integer specifying the total number of occupied cache buckets
	uint occupied;
	
	/// An array of pointers to objc_method data structures. This array may 
	/// contain no more than mask + 1 items. Note that pointers may be NULL, 
	/// indicating that the cache bucket is unoccupied, and occupied buckets 
	/// may not be contiguous. This array may grow over time
	Method buckets[0];
}

/// Specifies the superclass of an instance
struct objc_super {

	/// A pointer of type id. Specifies an instance of a class
	id receiver;
	
	/// A pointer to an objc_class data structure. Specifies the particular 
	/// superclass of the instance to message
	Class super_class;
}
