
// The D/Objective-C Bridge
// Copyright © 2007  Michel Fortin
// <http://michelf.com/d-objc-bridge/>
//
// This program is free software: you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free 
// Software Foundation, either version 3 of the License, or (at your option) 
// any later version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT 
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for 
// more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program. If not, see <http://www.gnu.org/licenses/>.

module objc.subclassinit;

import objc.runtime;
import objc.methodlist;
import objc.capsule;

import std.string;


//debug = ObjcBridgeLog;


/**
 * Objective-C subclass initializer template. Template argument subclassName 
 * contains the name of the subclass to create. The second argument, 
 * superClassName, contains the name of the Objective-C superclass and can 
 * be omitted if the parent D class also mixes-in this template.
 */
template ObjcSubclassInitializer(char[] subclassName, char[] superclassName = "") {

	// Assert className isn't empty.
	static assert(subclassName);
	
	
	/** Class name. */
	static const objcClassName = subclassName;


	/** Statically allocated class definition. */
	private static objc.runtime.objc_class objcClassDef;
	
	/** Statically allocated metaclass definition. */
	private static objc.runtime.objc_class objcMetaClassDef;


	/**
	 * Return a pointer to the created class data structure. If this class
	 * hasn't been added to the Objective-C runtime, it'll be added.
	 */
	static objc.runtime.Class objcClass()
	out (classRef) {
		assert(classRef);
	}
	body {
		if (objcClassDef.name == null) {
			objcClassInit;
		}
		return &objcClassDef;
	}


	/**
	 * Return a pointer to the created metaclass data structure. If this class
	 * hasn't been added to the Objective-C runtime, it'll be added.
	 */
	static objc.runtime.Class objcMetaClass()
	out (classRef) {
		assert(classRef);
	}
	body {
		if (objcMetaClassDef.name == null) {
			objcClassInit;
		}
		return &objcMetaClassDef;
	}
	
	
	/**
	 * Return a pointer to the superclass data structure. If the superclass
	 * hasn't been added to the Objective-C runtime, it'll be added.
	 */
	protected static objc.runtime.Class objcSuperClass()
	out (classRef) {
		assert(classRef);
	}
	body {
		static if (superclassName != "") {
			static objc.runtime.Class objcSuperClass;
			if (!objcSuperClass) {
				objcSuperClass = objc.runtime.objc_lookUpClass(superclassName.ptr);
			}
			return objcSuperClass;
		}
		else static if (is(typeof(&super.objcClass) == objc.runtime.Class function())) {
			return typeof(super).objcClass;
		}
		else {
			return objc.capsule.capsuleClass;
		}
	}
	
	
	/**
	 * Return a pointer to the created metaclass data structure. If this class
	 * hasn't been added to the Objective-C runtime, it'll be added.
	 */
	static objc.runtime.id objcPlaceholder() {
		static objc.runtime.id placeholder;
		if (placeholder == null) {
			placeholder = objcAlloc;
		}
		return placeholder;
	}
	
	
	/**
	 * List of instance method lists, including inherited methods 
	 * but excluding initializers (like [NSObject init]). This list is 
	 * used when creating a wrapper subclass which does not derive 
	 * directly from this class.
	 */
	protected static objc.runtime.objc_method_list*[] objcMethodLists;
	
	/**
	 * List of class methods, including inherited methods.
	 */
	protected static objc.runtime.objc_method_list*[] objcClassMethodLists;
	
	

	/**
	 * Invoke on this object the instance method with given name, return 
	 * type and arguemnts.
	 */
	protected template invokeObjcSelf(R, char[] name) {
		// The double-template trick (a function template within a template)
		// allows the arguments of callObjcSelf to be deduced, making the 
		// function call easier to write.
		final R invokeObjcSelf(A...)(A args) {
			alias objc.method.ObjcMethodSelectorCheck!(name, A) check;
			return objc.method.msgSend!(R, A)(context.receiver, objc.method.selector!(name), args);
		}
	}
	
	
	/**
	 * Invoke on this object the instance method with given name, return 
	 * type and arguemnts.
	 */
	protected template invokeObjcSuper(R, char[] name) {
		// The double-template trick (a function template within a template)
		// allows the arguments of callObjcSelf to be deduced, making the 
		// function call easier to write.
		final R invokeObjcSuper(A...)(A args) {
			alias objc.method.ObjcMethodSelectorCheck!(name, A) check;
			return objc.method.msgSendSuper!(R, A)(&context, objc.method.selector!(name), args);
		}
	}


	/**
	 * Invoke class method with given name on this class.
	 */
	protected template invokeObjcSelfClass(R, char[] name) {
		// The double-template trick (a function template within a template)
		// allows the arguments of callObjcSelf to be deduced, making the 
		// function call easier to write.
		static final R invokeObjcSelfClass(A...)(A args) {
			alias objc.method.ObjcMethodSelectorCheck!(name, A) check;
			objc.runtime.id selfClass = cast(objc.runtime.id)objcClass;
			assert(selfClass);
			return objc.method.msgSend!(R, A)(selfClass, objc.method.selector!(name), args);
		}
	}
	
	
	/**
	 * Invoke class method with given name on this superclass.
	 */
	protected template invokeObjcSuperClass(R, char[] name) {
		// The double-template trick (a function template within a template)
		// allows the arguments of callObjcSelf to be deduced, making the 
		// function call easier to write.
		static final R invokeObjcSuperClass(A...)(A args) {
			alias objc.method.ObjcMethodSelectorCheck!(name, A) check;
			objc.runtime.id superClass = cast(objc.runtime.id)objcSuperClass;
			assert(superClass);
			return objc.method.msgSend!(R, A)(superClass, objc.method.selector!(name), args);
		}
	}


	/**
	 * Normal Objective-C instance allocator function.
	 */
	static objc.runtime.id objcAlloc() {
		return objc.runtime.class_createInstance(objcClass, 0);
	}
	
	
	/**
	 * Instance allocator returning an incompletly-allocated placeholder object.
	 */
	static objc.runtime.id objcPlaceholderAlloc() {
		return objcPlaceholder;
	}
	/** ditto */
	static objc.runtime.id objcPlaceholderAllocWithZone(void* zone) {
		return objcPlaceholder;
	}
	
	/** Static allocation for the allocator method list. */
	private static objc.methodlist.objc_variable_method_list!(2) allocMethodListDef;
	
	/** Initialization of the allocator method list with placeholder allocators. */
	private static objc.runtime.objc_method_list* allocMethodList() {
		alias objc.runtime.id id;
		alias objc.runtime.SEL SEL;
		struct _NSZone { }; // struct with the right name for methodTypeEncode.
		
		with (allocMethodListDef.method_list[0]) {
			method_name  = objc.method.selector!("alloc");
			method_types = objc.typecoding.methodTypeEncode!(id, id, SEL).ptr;
			method_imp   = cast(objc.runtime.IMP)&objcPlaceholderAlloc;
		}
		with (allocMethodListDef.method_list[1]) {
			method_name  = objc.method.selector!("allocWithZone:");
			method_types = objc.typecoding.methodTypeEncode!(id, id, SEL, _NSZone*).ptr;
			method_imp   = cast(objc.runtime.IMP)&objcPlaceholderAllocWithZone;
		}

		return cast(objc.runtime.objc_method_list*)&allocMethodListDef;
	};
	

	/**
	 * Initializer that fill the class data structures and register it with
	 * the Objective-C runtime.
	 */
	private static void objcClassInit() {
	
		// Gather instance initializer method list.
		objc.runtime.objc_method_list* initMethodList = 
			objc.methodlist.objcMethodListOf!(typeof(this), objc.methodlist.MethodKind.INSTANCE_INIT);

		// Gather instance instance method list.
		objc.runtime.objc_method_list* methodList = 
			objc.methodlist.objcMethodListOf!(typeof(this), objc.methodlist.MethodKind.INSTANCE);
		
		if (methodList)  objcMethodLists ~= methodList;
		
		// Also append inherited wrapper method lists.
		static if (is(typeof(&super.objcClassDef == objc.runtime.Class))) {
			super.objcClass; // Make sure the superclass method lists are initialized.
			objcMethodLists ~= typeof(super).objcMethodLists;
		}
		
		// Gather class method list.
		objc.runtime.objc_method_list* classMethodList = 
			objc.methodlist.objcMethodListOf!(typeof(this), objc.methodlist.MethodKind.CLASS);
		
		if (classMethodList)  objcClassMethodLists ~= methodList;

		// Initialize this class.
		objc.subclassinit.subclassInit(subclassName.ptr, objcSuperClass, 
			&objcClassDef, &objcMetaClassDef,
			initMethodList ~ objcMethodLists, allocMethodList ~ objcClassMethodLists);
	}

}


// Check that this compiles.
private {
	static class objc_subclassinit_Test {
		mixin ObjcSubclassInitializer!(this.stringof, "NSObject");
	}
}


/**
 * Initialize subclass data and register it with the Objective-C runtime.
 */
void subclassInit(char* className, Class superClass, 
	Class selfClass, Class metaClass, 
	objc_method_list*[] instanceMethodLists = null, 
	objc_method_list*[] classMethodLists = null)
in {
	// Ensure that the superclass exists and that someone
	// hasn't already implemented a class with the same name
	assert(superClass);
	assert(selfClass);
	assert(metaClass);
	assert(!objc.runtime.objc_lookUpClass(className), 
		"Trying to initialize twice Objective-C class "~ std.string.toString(className) ~".");
}
body {
	// Find the root class
	objc.runtime.Class rootClass = superClass;
	while (rootClass.super_class != null) {
		rootClass = rootClass.super_class;
	}
 
	// Setup Class and objcMetaClass
	selfClass.isa  = metaClass;
	selfClass.info = objc.runtime.CLS_CLASS;
	metaClass.info = objc.runtime.CLS_META;
 
	// For efficiency, we have the objcMetaClass and the class itself 
	// to share this copy of the name, but this is not a requirement
	// imposed by the runtime.
	selfClass.name = className;
	metaClass.name = className;
 
	// Allocate empty method lists.
	// We can add methods later.
//	selfClass.methodLists = instanceMethodLists.ptr;
//	metaClass.methodLists = classMethodLists.ptr;
 
	// Connect the class definition to the class hierarchy:
	// Connect the class to the superclass.
	// Connect the objcMetaClass to the objcMetaClass of the superclass.
	// Connect the objcMetaClass of the objcMetaClass to the objcMetaClass of the root class.
	selfClass.super_class = superClass;
	metaClass.super_class = superClass.isa;
	metaClass.isa         = rootClass.isa;
 
	// Set the sizes of the class and the objcMetaClass.
	selfClass.instance_size = superClass.instance_size;
	metaClass.instance_size = metaClass.super_class.instance_size;
 
	// Finally, register the class with the runtime.
	objc.runtime.objc_addClass(selfClass);
	
	// And add class methods.
	foreach (methodList; instanceMethodLists) {
		objc.runtime.class_addMethods(selfClass, methodList);
	}
	foreach (methodList; classMethodLists) {
		objc.runtime.class_addMethods(metaClass, methodList);
	}
	
	debug (ObjcBridgeLog) {
		printf("Added Objective-C class %s\n", selfClass.name);
		
		// Show a sample of the instance methods.
		{
			alias instanceMethodLists methodLists;
		
			int count = methodLists.length ? methodLists[0].method_count : 0;
			for (uint i = 0; i < count; i++) {
				// Complicated trick to avoid GDC bug with static arrays in struct.
				with ((cast(objc_method*)(&instanceMethodLists[0].method_list))[i]) {
					printf("  -%s  '%s'\n", method_name, method_types);
				}
			}
			
			int additionalLists = methodLists.length-1;
			if (additionalLists == 1) {
				printf("+ 1 other instance method group.\n");
			} else if (additionalLists >= 2) {
				printf("+ %d other instance method groups.\n", additionalLists);
			}
		}
		
		// Show a sample of the class methods.
		{
			alias classMethodLists methodLists;
		
			int count = methodLists.length ? methodLists[0].method_count : 0;
			for (uint i = 0; i < count; i++) {
				// Complicated trick to avoid GDC bug with static arrays in struct.
				with ((cast(objc_method*)(&instanceMethodLists[0].method_list))[i]) {
					printf("  -%s  '%s'\n", method_name, method_types);
				}
			}
			
			int additionalLists = methodLists.length-1;
			if (additionalLists == 1) {
				printf("+ 1 other instance method group.\n");
			} else if (additionalLists >= 2) {
				printf("+ %d other instance method groups.\n", additionalLists);
			}
		}
	}
}

