
// The D/Objective-C Bridge
// Copyright © 2007  Michel Fortin
// <http://michelf.com/d-objc-bridge/>
//
// This program is free software: you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free 
// Software Foundation, either version 3 of the License, or (at your option) 
// any later version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT 
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for 
// more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program. If not, see <http://www.gnu.org/licenses/>.

/**
 * Templates for automatic type conversion between Objective-C and D.
 *
 * Most types require no conversion for passing through, but some need casts 
 * (like BOOL <-> bool) and some need encapsulation or wrapping (objects). 
 * This module provides a generic way to perform these conversions.
 */
module objc.type;

import objc.runtime;
import objc.typecoding;

import std.typetuple;


/**
 * Generic interface allowing a D class and an Objective-C class to be mapped 
 * together.
 */
interface ObjcObject {

	/**
	 * Get Objective-C object instance pointer.
	 */
	id objcInstance();

	/**
	 * Decapsulate Objective-C object instance into a suitable D object 
	 * descendent of ObjcObject.
	 */
	static ObjcObject objcWrapInstance(id self);

}


/**
 * Encapsulate D type T to corresponding Objective-C type.
 */
ObjcType!(T) encapsulate(T)(T value) {
	return value;
}

/**
 * Decapsulate D type T from corresponding Objective-C type.
 */
T decapsulate(T)(ObjcType!(T) value) {
	return value;
}


/**
 * Objective-C equivalent type for T.
 */
template ObjcType(T) {
	alias T ObjcType;
}


/**
 * Tuple of Objective-C equivalent types.
 */
template ObjcTypeTuple(T, U...) {
	alias TypeTuple!(ObjcType2!(T), ObjcTypeTuple!(U)) ObjcTypeTuple;
}


/**
 * Empty tuple of Objective-C equivalent types.
 */
template ObjcTypeTuple() {
	alias TypeTuple!() ObjcTypeTuple;
}


// -- void Specialization

void encapsulate(T : void)() {
}

void decapsulate(T : void)() {
}

template ObjcType(T : void) {
	alias void ObjcType;
}


// -- BOOL Specialization

BOOL encapsulate(T : bool)(bool var) {
	return cast(BOOL)var;
}

bool decapsulate(T : bool)(BOOL var) {
	return cast(bool)var;
}

template ObjcType(T : bool) {
	alias BOOL ObjcType;
}


// -- Generic Object Specialization using Capsule.

id encapsulate(T : object.Object)(T var) {
	static if (is(T : ObjcObject)) {
		return var ? var.objcInstance : null;
	} else {
		return encapsulateObjectFunc(var);
	}
}

T decapsulate(T : object.Object)(id var) {
	static if (is(T : ObjcObject)) {
		return T.objcWrapInstance(var);
	} else {
		return cast(T)decapsulateObjectFunc(var);
	}
}

template ObjcType(T : object.Object) {
	alias objc.runtime.id ObjcType;
}


static id function(object.Object) encapsulateObjectFunc;
static object.Object function(id) decapsulateObjectFunc;


/**
 * Ridiculous template front end for objc.type.ObjcType which avoids forward 
 * reference issues with class forward references.
 */
private template ObjcType2(T) {
	static if (is(T == class)) {
		alias id ObjcType2;
	}
	else {
		alias ObjcType!(T) ObjcType2;
	}
}



// Unit tests

unittest {
	// Instanciate types.
	assert(encapsulate!(int)(1) == 1);
	assert(decapsulate!(int)(-3) == -3);
	assert(encapsulate!(bool)(true) == YES);
	assert(decapsulate!(bool)(YES) == true);
	
	// Instanciation fails if not in unittest. (Compiler problem?)
	assert(encapsulate!(id)(null) == null);
	assert(decapsulate!(id)(null) == null);
	assert(encapsulate!(Class)(null) == null);
	assert(decapsulate!(Class)(null) == null);
	assert(encapsulate!(char*)(null) == null);
	assert(decapsulate!(char*)(null) == null);
	assert(encapsulate!(wchar*)(null) == null);
	assert(decapsulate!(wchar*)(null) == null);
	assert(encapsulate!(uint)(1) == 1);
	assert(decapsulate!(uint)(1) == 1);
	
	enum Test { A, B }
	assert(encapsulate!(Test)(Test.B) == 1);
	assert(decapsulate!(Test)(Test.B) == 1);
	
	// Check for ObjcObject.
	// ... missing.
}

