
// The D/Objective-C Bridge
// Copyright © 2007  Michel Fortin
// <http://michelf.com/d-objc-bridge/>
//
// This program is free software: you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free 
// Software Foundation, either version 3 of the License, or (at your option) 
// any later version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT 
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for 
// more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program. If not, see <http://www.gnu.org/licenses/>.

/**
 * Templates to create encoded type strings for the Objective-C runtime.
 */
module objc.typecoding;

import objc.runtime;


/**
 * Create the encoded (mangled) form of a type for the Objective-C runtime.
 */
// This one deals with type tuples.
template encode(T, U...) {
	const char[] encode = encode!(T) ~ encode!(U);
}

// This one deals with an empty type tuple.
template encode() {
	const char[] encode = "";
}


template encode(T : byte) {
	const char[] encode = "c";
}
template encode(T : ubyte) {
	const char[] encode = "C";
}
template encode(T : char) {
	const char[] encode = "C";
}

template encode(T : short) {
	const char[] encode = "s";
}
template encode(T : ushort) {
	const char[] encode = "S";
}
template encode(T : wchar) {
	const char[] encode = "S";
}

template encode(T : int) {
	const char[] encode = "l";
}
template encode(T : uint) {
	const char[] encode = "L";
}
template encode(T : dchar) {
	const char[] encode = "L";
}

template encode(T : long) {
	const char[] encode = "q";
}
template encode(T : ulong) {
	const char[] encode = "Q";
}

template encode(T : float) {
	const char[] encode = "f";
}
template encode(T : ifloat) {
	const char[] encode = "f";
}

template encode(T : double) {
	const char[] encode = "d";
}
template encode(T : idouble) {
	const char[] encode = "d";
}

template encode(T : bool) {
	const char[] encode = "B";
}
template encode(T : void) {
	const char[] encode = "v";
}
template encode(T : char*) {
	const char[] encode = "*";
}

template encode(T : id) {
	const char[] encode = "@";
}
template encode(T : Class) {
	const char[] encode = "#";
}
template encode(T : SEL) {
	const char[] encode = ":";
}


template encode(T : T*) {
	static if (is(T == function)) {
		const char[] encode = "^?";
	} else {
		const char[] encode = "^" ~ encode!(T);
	}
}
template encode(T : T**) {
	// Need to encode as an opaque type after a second level of indirection.
	const char[] encode = "^^" ~ opaqueEncode!(T);
}
template encode(T) {
	// Static array
	static if (T.mangleof[0] == 'G') {
		const char[] encode = "["
			~ ToString!(T.sizeof / ArrayType!(T).sizeof) 
			~ encode!(ArrayType!(T)) ~ "]";
	}
	// Dynamic array
	else static if (T.mangleof[0] == 'A') {
		const char[] encode = "{" 
			~ typename!(T) ~ "=" 
			~ encode!(size_t) ~ encode!(ArrayType!(T)*) ~ "}";
	}
	// Struct
	else static if (is(T == struct)) {
		const char[] encode = "{" 
			~ typename!(T) ~ "=" 
			~ encode!(typeof(T.tupleof)) ~ "}";
	}
	// Union
	else static if (is(T == union)) {
		const char[] encode = "(" 
			~ typename!(T) ~ "=" 
			~ encode!(typeof(T.tupleof)) ~ ")";
	}
	// Anything else is unknown
	else {
		const char[] encode = "?";
	}
}

// Array and unions should not expose their members after a second level of 
// indirection. This template deals with that.
private template opaqueEncode(T) {
	// Struct
	static if (is(T == struct)) {
		const char[] opaqueEncode = "{" ~ typename!(T) ~ "}";
	}
	// Union
	else static if (is(T == union)) {
		const char[] opaqueEncode = "(" ~ typename!(T) ~ ")";
	}
	// Anything else is encoded normally
	else {
		const char[] opaqueEncode = encode!(T);
	}
}

/** Template that gives a name to struct, union and dynamic array types. */
private template typename(T) {
	static if (is(T == struct) || is(T == union)) {
		static if (T.stringof[$-1] == ' ') {
			const char[] typename = T.stringof[0..$-1];
		} else {
			const char[] typename = T.stringof;
		}
	} else {
		const char[] typename = "D_" ~ T.mangleof;
	}
}

/** Template that gives the type of an array's content. */
private template ArrayType(A : A[]) {
	alias A ArrayType;
}


private {
	// Void
	static assert(encode!(void) == "v");
	
	// Integral types
	static assert(encode!(int) == "l");
	static assert(encode!(void*) == "^v");
	static assert(encode!(SEL) == ":");
	
	// Arrays
	static assert(encode!(char[0]) == "[0C]");
	static assert(encode!(char[]) == "{D_Aa=L*}");
	static assert(encode!(char[0][]) == "{D_AG0a=L^[0C]}");
	
	// Struct
	struct TestStruct { int a; char*[5] b; }
	static assert(encode!(TestStruct) == "{TestStruct=l[5*]}");
	static assert(encode!(TestStruct*) == "^{TestStruct=l[5*]}");
	static assert(encode!(TestStruct**) == "^^{TestStruct}");
	
	// Union
	union TestUnion { int a; char*[5] b; }
	static assert(encode!(TestUnion) == "(TestUnion=l[5*])");
	static assert(encode!(TestUnion*) == "^(TestUnion=l[5*])");
	static assert(encode!(TestUnion**) == "^^(TestUnion)");

	// Type tupples
	static assert(encode!(id, SEL, void*[0]) == "@:[0^v]");
}



version (ObjcMethodTypeNoOffsets)
{
	alias encode methodTypeEncode;
}
else
{
	import std.metastrings;

	/**
	 * Create the encoded (mangled) form of method arguments for the Objective-C 
	 * runtime.
	 *
	 * The runtime traditionally includes the stack size and the offset of 
	 * arguments as numbers interleaved between types. Those numbers are 
	 * unreliable and just get ignored by most recent versions of the runtime.
	 *
	 * Setting the ObjcMethodTypeNoOffsets version flag will make 
	 * `methodTypeEncode` an alias of `encode`, taking away those numbers.
	 */
	template methodTypeEncode(R, A...) {
		const char[] methodTypeEncode = 
			encode!(R) ~ ToString!(_sizeOfArgs!(A)) ~
			_methodArgumentsEncode!(0, A);
	}

	private template _methodArgumentsEncode(size_t offset, A, B...) {
		const char[] _methodArgumentsEncode =
			encode!(A) ~ ToString!(offset) ~
			_methodArgumentsEncode!(offset + ((A.sizeof + 3) / 4 * 4), B);
	}

	private template _methodArgumentsEncode(size_t offset) {
		const char[] _methodArgumentsEncode = "";
	}

	private template _sizeOfArgs(T, U...) {
		const size_t _sizeOfArgs = ((T.sizeof + 3) / 4 * 4) + _sizeOfArgs!(U);
	}

	private template _sizeOfArgs() {
		const size_t _sizeOfArgs = 0;
	}


	private {
		struct ArgStruct { uint a, b; }
		// Method type tuples
		static assert(methodTypeEncode!(void, id, SEL) == "v8@0:4");
		static assert(methodTypeEncode!(id, id, SEL, Class) == "@12@0:4#8");
		static assert(methodTypeEncode!(id, id, SEL, ArgStruct**, BOOL, short) == "@20@0:4^^{ArgStruct}8c12s16");
	}
}

