
// The D/Objective-C Bridge
// Copyright © 2007  Michel Fortin
// <http://michelf.com/d-objc-bridge/>
//
// This program is free software: you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free 
// Software Foundation, either version 3 of the License, or (at your option) 
// any later version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT 
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for 
// more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program. If not, see <http://www.gnu.org/licenses/>.

/**
 * Allow the creation of wrappers for Objective-C class. Wrappers can be 
 * constructed, used, deleted, and subsclassed as any D objects.
 */
module objc.wrapper;

import objc.runtime;
import objc.type;
import objc.objectlookup;


/**
 * Type for a function to wrap in Objective-C objects of a given class into
 * a proper D Object.
 */
alias Object function(id) WrapperFunc;



/**
 * Wrapper function to use when no wrapper was found for a given Objective-C 
 * class.
 */
WrapperFunc defaultWrapperFunc;


/**
 * Associative array for storing wrapper functions.
 */
private WrapperFunc[Class] wrapperLookup;


/**
 * Wrap object 'self' in an Object by looking up for a suitable 
 * wrapper function for its class. If no class-specific wrapper function
 * is found, return a vanilla Object.
 */
Object wrapInstance(id instance) {
	if (instance is null) return null;
	
	auto object = objc.objectlookup.lookup(instance);
	if (object) return object;

	WrapperFunc wrapF = wrapFunction(instance.isa);
	return wrapF(instance);
}


Object wrapInstanceNoLookup(id instance) 
in {
	assert(instance);
}
body {
	WrapperFunc wrapF = wrapFunction(instance.isa);
	return wrapF(instance);
}


WrapperFunc wrapFunction(Class objcClass) {
	WrapperFunc* funcPtr;

	do {
		funcPtr = (objcClass in wrapperLookup);
		if (funcPtr) return *funcPtr;
		
		objcClass = objcClass.super_class;
	} while (objcClass);
	
	return defaultWrapperFunc;
}


/**
 * Register an wrapper function for specified class. Registering twice
 * for the same class is an error.
 */
void registerWrapper(Class objcClass, WrapperFunc func) 
in {
//	assert((objcClass in wrapperLookup) == null);
}
body {
	wrapperLookup[objcClass] = func;
}


/**
 * Wrapper class for an Objective-C object.
 */
class ObjcWrapperBase : ObjcObject {
	
	/** Reference to Objective-C object instance method handler. */
	package objc_super context = { };
	
	invariant {
		assert(context.receiver);
		assert(context.super_class);

		// Wrong assertion: in some cases (NSString!) two different D objects
		// will have to point to the same Objective-C instance.
		//assert(lookup(context.receiver) is this);
	}


	/**
	 * Initialize object from an Objective-C object instance to wrap.
	 */
	protected this(id instance) {
		this.objcInstance = instance;
	}
	
	
	~this() {
		unregister(objcInstance);
	}


	/**
	 * Get Objective-C object instance pointer.
	 */
	id objcInstance() {
		return context.receiver;
	}
	
	/**
	 * Set Objective-C object instance pointer.
	 */
	protected void objcInstance(id instance) 
	in {
		assert(instance);
	}
	body {
		if (instance != context.receiver) {
			objc.objectlookup.unregister(context.receiver);
			context.receiver = instance;
			objc.objectlookup.register(context.receiver, this);
		}
		context.super_class = objcWrappedClassFor(instance.isa);
	}
	
	/**
	 * The superclass to call methods on for an instance of class c.
	 */
	protected Class objcWrappedClassFor(Class c) {
		return c;
	}
	
	
	/**
	 * Tell if the linked Objective-C instance is hard-wired to this specific
	 * D object and cannot exist without it. Only available in debug mode.
	 */
	debug protected bool objcHardWired() {
		return false;
	}
	
}


/**
 * Object wrapper methods to mixin within Object derivatives when
 * creating a wrapper for an Objective-C object.
 */
template ObjcWrapperDerived() {
	
	/** Type of the template host object. */
	private alias typeof(this) HostObject;
	
	/** Wrap an object instance within an object of this class. */
	private static Object objcCreateWrapper(objc.runtime.id instance) {
		// Retain the Objective-C instance prior wrapping it.
		// It needs to be done prior wrapping so that invariants holds
		// after the constructor is called.
		objc.method.msgSend!(void)(instance, objc.method.selector!("retain"));
		
		// Create new Wrapper.
		return new HostObject(instance);
	}
	
	/** Register encapsulator for wrapped Objective-C class.  */
	static void objcWrapInit() {
		objc.wrapper.registerWrapper(objcSuperClass, &objcCreateWrapper);
	}

	/**
	 * The superclass to call methods on for an instance of class c.
	 *
	 * If c is of the same class as our capsule subclass, we need to set it to 
	 * its superclass to avoid calling our methods recursivly. Otherwise we're 
	 * wrapping an Objective-C class in which case we should invoke methods 
	 * directly on that class.
	 */
	protected objc.runtime.Class objcWrappedClassFor(objc.runtime.Class c) {
		if (c == &objcClassDef) {
			c = objcSuperClass;
		}
		// If function is defined by super...
		static if (is(typeof(&super.objcWrappedClassFor) == objc.runtime.Class delegate(objc.runtime.Class))) {
			return super.objcWrappedClassFor(c);
		} else {
			return c;
		}
	}
	
//	/** Wrapping constructor */
//	mixin("this(objc.runtime.id instance) { super(instance); }");

}


